#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ludcmp.h"
#include "jacobn.c"
#define SAFETY 0.9
#define GROW 1.5
#define PGROW -0.25
#define SHRNK 0.5
#define PSHRNK (-1.0/3.0)
#define ERRCON 0.1296
#define MAXTRY 10000
#define GAM (1.0/2.0)
#define A21 2.0
#define A31 (48.0/25.0)
#define A32 (6.0/25.0)
#define C21 -8.0
#define C31 (372.0/25.0)
#define C32 (12.0/5.0)
#define C41 (-112.0/125.0)
#define C42 (-54.0/125.0)
#define C43 (-2.0/5.0)
#define B1 (19.0/9.0)
#define B2 (1.0/2.0)
#define B3 (25.0/108.0)
#define B4 (125.0/108.0)
#define E1 (17.0/54.0)
#define E2 (7.0/36.0)
#define E3 0.0
#define E4 (125.0/108.0)
#define C1X (1.0/2.0)
#define C2X (-3.0/2.0)
#define C3X (121.0/50.0)
#define C4X (29.0/250.0)
#define A2X 1.0
#define A3X (3.0/5.0)
float my_fmax(float a, float b) {
	if (a>0) 
		return a;
	else
		return b;
}
float my_fmin(float a, float b) {
	if (a<0) 
		return a;
	else
		return b;
}
/*Vrati pole derivaci podle soustavy ODR*/
void derivs(float x, float y[], float dydx[]) {
	/*Rychlosti reakci*/
	float k1=1e-31;
	float k2=1e-9;
	float k3=1e-7;
	float k4=100;
	float k5=1e-8;
	float He=1e17;
	/*Soustava ODR*/
	dydx[0]=k1*y[4]*He*He-k3*y[0]*y[3];
	dydx[1]=-k2*y[1]*y[1];
	dydx[2]=-k5*y[3]*y[2]+k4*y[4];
	dydx[3]=-k5*y[3]*y[2]+k2*y[1]*y[1]-k3*y[0]*y[3];
	dydx[4]=-k1*y[4]*He*He+k2*y[1]*y[1]-k4*y[4];
}

void stiff(float y[],float dydx[], int n, float *x, float htry, float eps, float yscal[], float *hdid, float *hnext, void (*derivs)(float, float [], float [])) {
	int i, j, jtry, *indx;
	float d, errmax, h, xsav, **a, *dfdx, **dfdy, *dysav, *err;
	float *g1, *g2, *g3, *g4, *ysav;
	/*Alokace promennych*/
	indx=(int *)malloc(n*sizeof(int));

	a=malloc(n*sizeof(float *));
	for (i=0; i<n; i++)
		a[i]=(float *)malloc(n*sizeof(float));
	dfdx=(float *)malloc(n*sizeof(float));

	dfdy=malloc(n*sizeof(float *));
	for (i=0; i<n; i++)
		dfdy[i]=(float *)malloc(n*sizeof(float));
	
	dysav=malloc(n*sizeof(float *));
	err=malloc(n*sizeof(float *));
	g1=malloc(n*sizeof(float *));
	g2=malloc(n*sizeof(float *));
	g3=malloc(n*sizeof(float *));
	g4=malloc(n*sizeof(float *));
	ysav=malloc(n*sizeof(float *));
	xsav=(*x);
	/*Ulozeni vstupnich dat*/
	for (i=0; i<n; i++) {
		ysav[i]=y[i];
		dysav[i]=dydx[i];
	}
	jacobn(xsav, ysav, dfdx, dfdy, n);
	h=htry;
	for (jtry=1; jtry<=MAXTRY; jtry++) {
		for (i=0; i<n; i++) {
			for (j=0; j<n; j++)
				a[i][j]=-dfdy[i][j];
			a[i][i]+=1.0/(GAM*h);
		}
		ludcmp(a,n,indx,&d);
		for (i=0; i<n; i++)
			g1[i]=dysav[i]+h*C1X*dfdx[i];
		lubksb(a,n,indx,g1);
		for (i=0; i<n; i++)
			y[i]=ysav[i]+A21*g1[i];
		*x=xsav+A2X*h;
		(*derivs)(*x,y,dydx);
		for (i=0; i<n; i++)
			g2[i]=dydx[i]+h*C2X*dfdx[i]+C21*g1[i]/h;
		lubksb(a,n,indx,g2);
		for (i=0; i<n; i++)
			y[i]=ysav[i]+A31*g1[i]+A32*g2[i];
		*x=xsav+A3X*h;
		(*derivs)(*x,y,dydx);
		for (i=0; i<n; i++)
			g3[i]=dydx[i]+h*C3X*dfdx[i]+(C31*g1[i]+C32*g2[i])/h;
		lubksb(a,n,indx,g3);
		for (i=0; i<n; i++)
			g4[i]=dydx[i]+h*C4X*dfdx[i]+(C41*g1[i]+C42*g2[i]+C43*g3[i])/h;
		lubksb(a,n,indx,g4);
		for (i=0; i<n; i++) {
			y[i]=ysav[i]+B1*g1[i]+B2*g2[i]+B3*g3[i]+B4*g4[i];
			err[i]=E1*g1[i]+E2*g2[i]+E3*g3[i]+E4*g4[i];
		}
		*x=xsav+h;
		if (*x==xsav) {
			printf("Chyba kroku ve stiff.");
			exit(0);
		}
		errmax=0.0;
		for (i=0; i<n; i++) 
			if (errmax<fabs(err[i]/yscal[i]))
				errmax=fabs(err[i]/yscal[i]);
		errmax/=eps;
		if (errmax <=1.0) {
			*hdid=h;
			if (errmax>ERRCON)
				*hnext=SAFETY*h*pow(errmax,PGROW);
			else *hnext=GROW*h;
			free(ysav);
			free(g4);
			free(g3);
			free(g2);
			free(g1);
			free(err);
			free(dysav);
			for (i=0; i<n; i++)
				free(dfdy[i]);
			free(dfdy);
			free(dfdx);
			for (i=0; i<n; i++)
				free(a[i]);
			free(a);
			free(indx);
			return;
		} else {
			*hnext=SAFETY*h*pow(errmax,PSHRNK);
			h=(h>=0.0 ? my_fmax(*hnext, SHRNK*h) : my_fmin(*hnext, SHRNK*h));
		}
	}
	printf("Chyba: Prekrocen maximalni pocet iteraci.\n");
	exit(1);
}

/*Hlavni metoda, vlastni vypocet*/
int main(int argc, char *argv[]) {
	int i;
	float dydx[5];
	float *yscal;
	float y[5];
	float *hdid, *hnext;
	float *t;
	/*Pocatecni koncetrace*/
	y[0]=0;
	y[1]=1e10;
	y[2]=0;
	y[3]=1e10;
	y[4]=1e10;
	dydx[0]=0;
	dydx[1]=0;
	dydx[2]=0;
	dydx[3]=0;
	dydx[4]=0;
	t=malloc(sizeof(float));
	hdid=malloc(sizeof(float));
	hnext=malloc(sizeof(float));
	yscal=(float *)malloc(5*sizeof(float));
	yscal[0]=1;
	yscal[1]=1;
	yscal[2]=1;
	yscal[3]=1;
	yscal[4]=1;
	*t=0;
	i=0;
	while (*t<10) {
//		if (i%100==0)
			printf("%e\t%e\t%e\t%e\t%e\t%e\n",*t,y[0],y[1],y[2],y[3],y[4]);
		stiff(y,dydx, 5, t, 1e-3, 1000, yscal, hdid, hnext, &derivs);
		i++;
	}
	fprintf(stderr,"Pocet kroku: %d\n",i);
}
